#include "extconf.h"

#include <stdarg.h>

#if defined(_WIN32) || defined(__WIN32__) || defined(WIN32)
# include <windows.h>
# ifndef __WIN32__
#  define __WIN32__
# endif
#else
# ifndef _GNU_SOURCE
#  define _GNU_SOURCE
# endif
# include <unistd.h>
# include <sys/mman.h>
# ifndef MAP_ANONYMOUS
#  define MAP_ANONYMOUS MAP_ANON
# endif
#endif

#include <string>
#include <map>
#include <set>
#include <iostream>

using namespace std;

extern "C" {

#include <ruby/ruby.h>
#include <ruby/node.h>

#include "vm_core.h"

// Generated header
#include "insns.inc"

//#define DEBUG
#ifdef DEBUG
# define LOGL(line, msg, ...) do {                            \
        printf("YJ(%d): "msg, line, ## __VA_ARGS__);          \
        fflush(stdout);                                       \
    } while (0)
#else
# define LOGL(line, msg, ...) do {} while (0)
#endif

#define LOG(msg, ...) LOGL(__LINE__, msg, ## __VA_ARGS__)

#if defined(__i386__)
# define SELF_REG esi
# define SP esp
# define BP ebp
# define AX eax
# define BX ebx
# define CX ecx
# define DX edx
# define ARCH "i386"

#elif defined(__x86_64__)
# define SELF_REG rbx
# define SP rsp
# define BP rbp
# define AX rax
# define BX rbx
# define CX rcx
# define DX rdx
# define ARCH "i386:x86-64"
#else
# error "Unspported architecture"
#endif

#if defined(__APPLE__)
// Late binder of dyld assumes that the stack pointer is aligned for SSE.
# define YAJIT_CALL __attribute__((force_align_arg_pointer))
#else
# define YAJIT_CALL
#endif

#define PTRSZ(p) (sizeof(void*)*(p))

static VALUE rb_yajit_func_call(int argc, VALUE* argv,
                                VALUE jitval, VALUE self);
static VALUE yajit_func_call(int argc, VALUE* argv, VALUE self);

static ptrdiff_t rarray_len_offset;
static ptrdiff_t rarray_ptr_offset;

typedef unsigned long rb_num_t;
typedef VALUE(*ruby_anyarg_func_t)(ANYARGS);

extern NODE *rb_method_body(VALUE body);
extern VALUE rb_reg_new_ary(VALUE ary, int opt);
extern VALUE rb_vm_top_self();

}

static int pagesize = 0;

static volatile void* df[512];
static volatile void** dfp = df;
static volatile void* lf[512];
static volatile void** lfpp = lf;
static volatile void** lfp = 0;

struct rb_jit_func_t {
    VALUE recv;
    void* code;
    void* top;
    int size;
    int stack_max;
    int local_size;
    bool is_block;
    int argc;
    VALUE name;
    VALUE children;
    bool need_protect;

    void makeExecutable() {
        if (!need_protect) return;
#ifdef __WIN32__
        need_protect = false;
        DWORD oldProtect;
        if (!VirtualProtect(const_cast<void*>(code), size,
                            PAGE_EXECUTE_READWRITE, &oldProtect)) {
            fprintf(stderr, "can't protect (code=%p, size=%d, canExec=%d)\n",
                    code, size, 1);
            exit(1);
        }
#else
        char* aligned_addr = (char*)((size_t)code & ~(pagesize-1ULL));
        size_t padding = (char*)code - aligned_addr;
        int aligned_size = pagesize + (size + padding) & ~(pagesize-1);

        if (mprotect(aligned_addr,
                     aligned_size,
                     PROT_READ | PROT_WRITE | PROT_EXEC) < 0) {
            fprintf(stderr, "YJ: mprotect");
            exit(1);
        }
#endif
    }
};

YAJIT_CALL
static void yajit_p(VALUE v, const char* msg = NULL) {
    if (msg) {
        LOG("%s: %s\n", msg, RSTRING_PTR(rb_inspect(v)));
    }
    else {
        LOG("%s\n", RSTRING_PTR(rb_inspect(v)));
    }
}

YAJIT_CALL
static void yajit_p_tmp(VALUE v, const char* msg = NULL) {
    if (msg) {
        printf("%s: %s\n", msg, RSTRING_PTR(rb_inspect(v)));
    }
    else {
        printf("%s\n", RSTRING_PTR(rb_inspect(v)));
    }
}

#define GetYajitFuncPtr(obj, ptr) \
    GetCoreDataFromValue(obj, rb_jit_func_t, ptr)

static void jit_func_mark(void* ptr) {
    if (ptr) {
        // TODO
        rb_jit_func_t* jit = (rb_jit_func_t*)ptr;
        rb_gc_mark(jit->name);
        rb_gc_mark(jit->children);
        rb_gc_mark(jit->recv);
    }
}

static void jit_func_free(void* ptr) {
    // TODO
}

#include "xbyak/xbyak.h"

static VALUE rb_cYajitFunc;

static VALUE vm;
static VALUE iseq;
static VALUE insn_name_array;

static const char* get_insn_name(int insn) {
    return RSTRING_PTR(RARRAY_PTR(insn_name_array)[insn]);
}

class Yajit : public Xbyak::CodeGenerator {
public:
    static void init() {
//#ifdef HAVE_MREMAP
#ifndef __WIN32__
        assert(pagesize);
        mapped_size = pagesize;
        while (used_size > mapped_size) {
            mapped_size *= 2;
        }
        mapped = (char*)mmap(NULL, Yajit::mapped_size,
                             PROT_READ | PROT_WRITE | PROT_EXEC,
                             MAP_PRIVATE | MAP_ANONYMOUS, -1, 0);
        if (!mapped) {
            puts("No memory?");
            exit(0);
        }
#endif
    }

    Yajit(const string& name0, bool is_block0, rb_iseq_t* iseq0, int depth0 = 0)
        : Xbyak::CodeGenerator(40960),
          name(name0), is_block(is_block0),
          iseq(iseq0), adjust_vma(0), uniq_id(0), uniq_id2(next_uniq_id2++),
          depth(depth0), trampoline_addr(NULL) {
        children = rb_ary_new();
    }

    char* genLabel(int v) {
        static char buf[256];
        sprintf(buf, "L%d_%d", uniq_id2, v);
        return buf;
    }

    char* genLabel() {
        static char buf[256];
        sprintf(buf, "l%d", uniq_id++);
        return buf;
    }

    template <typename Ptr>
    void call_abs(Ptr p) {
        LOG("call_abs: %p\n", p);
        mov(AX, (Xbyak::uint64)p);
        call(AX);
    }

    void adjust_arg_stack(int n, int n64 = 0) {
#if defined(__x86_64__)
        if (n64) {
            add(SP, PTRSZ(n64));
        }
#else
        add(SP, PTRSZ(n));
#endif
    }

    template <typename A>
    void arg1(A a) {
#if defined(__x86_64__)
        mov(rdi, a);
#else
        push(a);
#endif
    }

    template <typename A>
    void arg2(A a) {
#if defined(__x86_64__)
        mov(rsi, a);
#else
        push(a);
#endif
    }

    template <typename A>
    void arg3(A a) {
#if defined(__x86_64__)
        mov(rdx, a);
#else
        push(a);
#endif
    }

    template <typename A>
    void arg4(A a) {
#if defined(__x86_64__)
        mov(rcx, a);
#else
        push(a);
#endif
    }

    template <typename A>
    void arg5(A a) {
#if defined(__x86_64__)
        mov(r8, a);
#else
        push(a);
#endif
    }

    template <typename A>
    void arg6(A a) {
#if defined(__x86_64__)
        mov(r9, a);
#else
        push(a);
#endif
    }

    template <typename FP, typename A1>
    void call1(FP p, A1 a1) {
        arg1(a1);
        call_abs(p);
        adjust_arg_stack(1);
    }

    void send(ID opt_id, rb_num_t op_argc,
              rb_iseq_t* blockiseq, rb_num_t op_flag, IC ic);

    VALUE compile(bool need_trampoline = false);

    void compile_iseq(rb_iseq_t* iseq, bool need_leave = true);

    ruby_anyarg_func_t get_trampoline() {
        assert(trampoline_addr);
        return (ruby_anyarg_func_t)trampoline_addr;
    }

    void create_trampoline(VALUE jitval);

    void dump();

    void pop(const Xbyak::Operand& op) {
        Xbyak::CodeGenerator::pop(op);
    }

    void push(const Xbyak::Operand& op) {
        Xbyak::CodeGenerator::push(op);
    }

    void push(Xbyak::uint32 imm) {
        Xbyak::CodeGenerator::push(imm);
    }

    void swap_stack(int a, int b) {
        LOG("swap_stack: %d %d\n", a, b);
        mov(AX, ptr[SP + PTRSZ(a)]);
        mov(DX, ptr[SP + PTRSZ(b)]);
        mov(ptr[SP + PTRSZ(a)], DX);
        mov(ptr[SP + PTRSZ(b)], AX);
    }

    void load_const_addr(void* p) {
        // TODO: fix
#if defined(__x86_64__)
        mov(AX, ptr[(Xbyak::uint64)p]);
#else
        mov(AX, (Xbyak::uint64)p);
        mov(AX, ptr[AX]);
#endif
    }

    void rb_jit_func_new(rb_jit_func_t* jit_func);

    string name;
    bool is_block;
    rb_iseq_t* iseq;
    void* adjust_vma;
    int uniq_id;
    int uniq_id2;
    VALUE children;
    int depth;
    char* trampoline_addr;

    static char* mapped;
    static size_t used_size;
    static size_t mapped_size;
    static int next_uniq_id2;
};

char* Yajit::mapped = NULL;
size_t Yajit::used_size = 0;
size_t Yajit::mapped_size = 4096;
int Yajit::next_uniq_id2 = 0;

extern "C" {
    VALUE iseq_data_to_ary(rb_iseq_t *iseq);
}

void Yajit::rb_jit_func_new(rb_jit_func_t* jit_func) {
    const void* code = getCode();
    int size = getSize();
    jit_func->need_protect = true;

#ifndef __WIN32__
    jit_func->need_protect = false;
    size_t next_used_size = 16 + used_size + size & ~15;
    if (next_used_size > mapped_size) {
        size_t orig_mapped_size = mapped_size;
        while (next_used_size > mapped_size) {
            mapped_size *= 2;
        }
#ifdef HAVE_MREMAP
        void* next_mapped = mremap(mapped, orig_mapped_size, mapped_size, 0);
#else
        char* next_mapped = (char*)mmap(mapped + orig_mapped_size,
                                        mapped_size - orig_mapped_size,
                                        PROT_READ | PROT_WRITE | PROT_EXEC,
                                        MAP_PRIVATE | MAP_ANONYMOUS,
                                        -1, 0);
#endif
        if (next_mapped == (char*)-1) {
            LOG("mremap failed, acquire new page\n");
            used_size = 16 + size & ~15;
            init();
        }
#ifdef HAVE_MREMAP
        //assert(mapped == next_mapped);
        LOG("mremap: %p=>%p @%p\n",
            orig_mapped_size, mapped_size, mapped);
#else
        if (mapped != next_mapped - orig_mapped_size) {
            used_size = 16 + size & ~15;
            mapped = next_mapped;
            mapped_size -= orig_mapped_size;
            LOG("mmap other: %p=>%p @%p\n",
                orig_mapped_size, mapped_size, mapped);
        }
        else {
            LOG("mmap: %p=>%p @%p %p\n",
                orig_mapped_size, mapped_size, mapped, next_mapped);
        }
#endif
    }
    jit_func->top = NULL;
    jit_func->code = mapped + used_size;
    LOG("%p\n", jit_func->code);
    ((char*)jit_func->code)[0] = 'X';
    used_size = next_used_size;
#else
    jit_func->top = malloc(size+16);
    jit_func->code = (void*)(16 + (size_t)jit_func->top & ~15);
#endif
    memcpy(jit_func->code, code, size);
    adjust_vma = jit_func->code;
    jit_func->size = size;
    jit_func->stack_max = iseq->stack_max;
    jit_func->local_size = iseq->local_size;
    jit_func->is_block = is_block;
    jit_func->argc = iseq->argc;
    jit_func->name = rb_str_new2(name.c_str());
    jit_func->children = children;
    jit_func->recv = rb_vm_top_self();
    //jit_func->recv = Qnil;
    //jit_func->makeExecutable();
}

YAJIT_CALL
static VALUE yajit_call_block(VALUE self, VALUE block_func,
                              int argc, VALUE* argv) {
    return yajit_func_call(argc, argv, block_func);
}


YAJIT_CALL
static VALUE rb_funcall2_rev(VALUE self, ID mid, int argc, VALUE* argv) {
#if defined(__APPLE__) && defined(__x86_64__)
    alloca(16);
    asm volatile("and $-16, %%rsp;\n"
                 "add $16, %%rsp;\n"
                 :::"%rsp");
#endif
    LOG("rb_funcall2_rev: self=%p mid=%d argc=%d\n",
        self, mid, argc);
    int argc_2 = argc / 2;
    for (int i = 0; i < argc_2; i++) {
        VALUE tmp = argv[i];
        argv[i] = argv[argc-i-1];
        argv[argc-i-1] = tmp;
    }
    rb_gc_enable();
    return rb_funcall2(self, mid, argc, argv);
}

YAJIT_CALL
static VALUE rb_block_call_rev(VALUE self, ID mid, int argc, VALUE* argv,
                               VALUE(*bl_proc)(ANYARGS), VALUE data2 ) {
#if defined(__APPLE__) && defined(__x86_64__)
    alloca(16);
    asm volatile("and $-16, %%rsp;\n"
                 "add $16, %%rsp;\n"
                 :::"%rsp");
#endif
    LOG("rb_block_call_rev: self=%p mid=%d argc=%d bl_proc=%p data2=%p\n",
        self, mid, argc, bl_proc, data2);
    int argc_2 = argc / 2;
    for (int i = 0; i < argc_2; i++) {
        VALUE tmp = argv[i];
        argv[i] = argv[argc-i-1];
        argv[argc-i-1] = tmp;
    }
    rb_gc_enable();
    return rb_block_call(self, mid, argc, argv, bl_proc, data2);
}

void Yajit::send(ID op_id, rb_num_t op_argc,
                 rb_iseq_t* blockiseq, rb_num_t op_flag, IC ic) {
    int argc = (int)op_argc;

    VALUE block_func;
    if (blockiseq) {
        LOG("blockiseq=%d argc=%d\n", blockiseq, argc);
        Yajit block_jit(name+"(block)", true, blockiseq, depth+1);
        block_func = block_jit.compile();
        rb_ary_push(children, block_func);
#ifdef DEBUG
        block_jit.dump();
#endif
        arg6(block_func);
        arg5((Xbyak::uint64)&yajit_call_block);
    }

#if defined(__x86_64__)
    arg4(SP);
#else
    if (blockiseq) {
        mov(AX, SP);
        add(AX, PTRSZ(2));
        push(AX);
    }
    else {
        arg4(SP);
    }
#endif
    arg3(argc);
    arg2(op_id);
    // self
    if (op_flag & VM_CALL_FCALL_BIT) {
        arg1(SELF_REG);
    }
    else {
#if defined(__x86_64__)
        arg1(ptr[SP + PTRSZ(argc)]);
#else
        arg1(ptr[SP + PTRSZ(argc + (blockiseq ? 5 : 3))]);
#endif
    }

    if (blockiseq) {
        call_abs(&rb_block_call_rev);
    }
    else {
        //xor(rax, rax);
        call_abs(&rb_funcall2_rev);
    }

#if defined(__x86_64__)
    add(SP, PTRSZ(argc));
#else
    add(SP, PTRSZ(argc + (blockiseq ? 6 : 4)));
#endif
    mov(ptr[SP], AX);
}

#if defined(__x86_64__)

# define NORMAL_DISPATCH0(id) do {               \
        extern ID id;                            \
        pop(rdi);                                \
        mov(rsi, id);                            \
        mov(rdx, 0);                             \
        call_abs(&rb_funcall2);                  \
        push(AX);                                \
    } while(0); break

# define NORMAL_DISPATCH1(id) do {               \
        extern ID id;                            \
        mov(rcx, SP);                            \
        mov(rdi, ptr[SP + PTRSZ(1)]);            \
        mov(rsi, id);                            \
        mov(rdx, 1);                             \
        call_abs(&rb_funcall2);                  \
        add(SP, PTRSZ(2));                       \
        push(AX);                                \
    } while(0); break

#else

# define NORMAL_DISPATCH0(id) do {               \
        extern ID id;                            \
        pop(eax);                                \
        push(0);                                 \
        push(id);                                \
        push(eax);                               \
        call_abs(&rb_funcall2);                  \
        add(SP, PTRSZ(3));                       \
        push(AX);                                \
    } while(0); break

# define NORMAL_DISPATCH1(id) do {               \
        extern ID id;                            \
        push(SP);                                \
        push(1);                                 \
        push(id);                                \
        push(ptr[SP + PTRSZ(4)]);                \
        call_abs(&rb_funcall2);                  \
        add(SP, PTRSZ(6));                       \
        push(AX);                                \
    } while(0); break

#endif

YAJIT_CALL
static VALUE rb_ary_new_rev(long n, const VALUE* elts) {
    int i;
    VALUE ary;

    ary = rb_ary_new2(n);
    for (i = 0; i < n; i++) {
        RARRAY_PTR(ary)[i] = elts[n - i - 1];
    }
    RARRAY(ary)->len = n;

    return ary;
}

YAJIT_CALL
static VALUE yajit_const_get(VALUE self, ID id, VALUE top) {
    // TODO: apparently not correct implementation.
    yajit_p(self, "yajit_const_get self");
    yajit_p(id, "yajit_const_get id");
    yajit_p(top, "yajit_const_get top");
    VALUE v = rb_const_get(CLASS_OF(self), id);
    yajit_p(v, "yajit_const_get ret");
    return v;
}

YAJIT_CALL
static void yajit_define_method(const char* name, ruby_anyarg_func_t fp,
                                rb_num_t is_singleton, VALUE self, VALUE obj) {
    rb_gc_enable();
    VALUE klass;
    if (is_singleton) {
        if (FIXNUM_P(obj) || SYMBOL_P(obj)) {
            rb_raise(rb_eTypeError,
                     "can't define singleton method \"%s\" for %s",
                     name, rb_obj_classname(obj));
        }
        if (OBJ_FROZEN(obj)) {
            rb_error_frozen("object");
        }
        klass = rb_singleton_class(obj);
    }
    else {
        if (TYPE(self) == T_CLASS) {
            klass = self;
        }
        else {
            klass = CLASS_OF(self);
        }
    }
    LOG("DEF METHOD: name=%s fp=%p s=%d klass=%s self=%s obj=%s\n",
        name, fp, is_singleton,
        RSTRING_PTR(rb_obj_as_string(klass)),
        RSTRING_PTR(rb_obj_as_string(self)),
        RSTRING_PTR(rb_obj_as_string(obj)));
    rb_define_method(klass, name, fp, -1);
}

YAJIT_CALL
static VALUE yajit_define_class(ID id, rb_num_t define_type,
                                VALUE super, VALUE cbase) {
    rb_gc_enable();
    if (NIL_P(super)) {
        super = rb_cObject;
    }
    VALUE klass = rb_define_class(rb_id2name(id), super);
    return klass;
}

YAJIT_CALL
static VALUE yajit_check_include_array(VALUE flag, VALUE ary, VALUE obj) {
    extern ID idEqq;
    VALUE result;

    // From Ruby's code
    int i;
    result = Qfalse;

    if (TYPE(ary) != T_ARRAY) {
        ary = rb_Array(ary);
    }

    if (flag == Qtrue) {
        /* NODE_CASE */
        for (i = 0; i < RARRAY_LEN(ary); i++) {
            /* TODO: fix me (use another method dispatch) */
            if (RTEST(rb_funcall2(RARRAY_PTR(ary)[i], idEqq, 1, &obj))) {
                result = Qtrue;
                break;
            }
        }
    }
    else {
        obj = Qfalse;
        /* NODE_WHEN */
        for (i = 0; i < RARRAY_LEN(ary); i++) {
            if (RTEST(RARRAY_PTR(ary)[i])) {
                obj = result = Qtrue;
                break;
            }
        }
    }
    return result;
}

void Yajit::compile_iseq(rb_iseq_t* iseq, bool need_leave) {
    for (int i = 0; i < iseq->iseq_size; ) {
        L(genLabel(i));

        int insn = iseq->iseq[i];
        LOG("%d %s\n", i, get_insn_name(insn));
        switch (insn) {
        case BIN(send): {
            send(iseq->iseq[i+1], iseq->iseq[i+2], (rb_iseq_t*)iseq->iseq[i+3],
                 iseq->iseq[i+4], (IC)iseq->iseq[i+5]);
            break;
        }
#ifdef HAVE_putcbase
        case BIN(putcbase): {
            // TODO:
            push(rb_cObject);
            break;
        }
#endif
        case BIN(tostring): {
#if defined(__x86_64__)
            pop(rdi);
            call_abs(&rb_obj_as_string);
            push(AX);
#else
            call_abs(&rb_obj_as_string);
            mov(ptr[SP], AX);
#endif
            break;
        }
        case BIN(toregexp): {
            int opt = iseq->iseq[i+1];
#if RUBY_REVISION < 15000 && RUBY_REVISION > 0
            // for ruby-1.9.0-0
#if defined(__x86_64__)
            pop(rdi);
            mov(rsi, opt);
#else
            push(opt);
            push(ptr[SP + PTRSZ(1)]);
#endif
            call_abs(&rb_reg_new);
            adjust_arg_stack(3);
            push(AX);

            i--;
#else
            // for ruby-1.9.0-1 or later
            int num = iseq->iseq[i+2];
#if defined(__x86_64__)
            mov(rdi, num);
            mov(rsi, SP);
#else
            push(SP);
            push(num);
#endif
            call_abs(&rb_ary_new_rev);
#if defined(__x86_64__)
            mov(rdi, rax);
            mov(rsi, opt);
#else
            push(opt);
            push(eax);
#endif
            call_abs(&rb_reg_new_ary);
            adjust_arg_stack(num + 4, num);
#endif
            push(AX);
            break;
        }
        case BIN(newarray): {
            int num = iseq->iseq[i+1];
            if (num) {
                arg2(SP);
                arg1(num);
                call_abs(&rb_ary_new_rev);
                adjust_arg_stack(num + 2, num);
                push(AX);
            }
            else {
                call_abs(&rb_ary_new);
                push(AX);
            }
            break;
        }
        case BIN(newhash): {
            int num = iseq->iseq[i+1];
            assert((num & 1) == 0);
            call_abs(&rb_hash_new);
            push(AX);
            if (num == 0) {
                break;
            }

            for (int i = num; i > 0; i -= 2) {
#if defined(__x86_64__)
                mov(rdi, ptr[SP]);
                mov(rsi, ptr[SP + PTRSZ(i)]);
                mov(rdx, ptr[SP + PTRSZ(i - 1)]);
#else
                if (i < num) {
                    mov(eax, ptr[SP]);
                    add(SP, PTRSZ(2));
                    mov(ptr[SP], eax);
                }
                swap_stack(1, 2);
#endif
                call_abs(&rb_hash_aset);
            }
            pop(AX);
            adjust_arg_stack(2, num);
            push(AX);
            break;
        }
        case BIN(concatstrings): {
            int num = iseq->iseq[i+1];
#if defined(__x86_64__)
            xor(rdi, rdi);
            xor(rsi, rsi);
#else
            push(0);
            push(0);
#endif
            call_abs(&rb_str_new);
            for (int i = 0; i < num; i++) {
#if defined(__x86_64__)
                mov(rdi, rax);
                mov(rsi, ptr[SP + PTRSZ(num - i - 1)]);
#else
                mov(ptr[SP], eax);
                mov(eax, ptr[SP + PTRSZ(num - i + 1)]);
                mov(ptr[SP + PTRSZ(1)], eax);
#endif
                call_abs(&rb_str_append);
            }
            adjust_arg_stack(num + 2, num);
            push(AX);
            break;
        }
        case BIN(newrange): {
#if defined(__x86_64__)
            mov(rdx, iseq->iseq[i+1]);
            pop(rsi);
            pop(rdi);
#else
            push(iseq->iseq[i+1]);
            push(ptr[SP + PTRSZ(1)]);
            push(ptr[SP + PTRSZ(3)]);
#endif
            call_abs(&rb_range_new);
            adjust_arg_stack(5);
            push(AX);
            break;
        }
        case BIN(putself): {
            push(SELF_REG);
            break;
        }
        case BIN(putstring): {
            call1(&rb_str_new3, iseq->iseq[i+1]);
            push(AX);
            break;
        }
        case BIN(duparray): {
            call1(&rb_ary_dup, iseq->iseq[i+1]);
            push(AX);
            break;
        }
        case BIN(putobject): {
            VALUE obj = iseq->iseq[i+1];
#if defined(__x86_64__)
            if (obj < (1LL << 32)) {
                push(obj);
            }
            else {
                mov(rax, obj);
                push(rax);
            }
#else
            push(obj);
#endif
            break;
        }
        case BIN(checkincludearray): {
            VALUE flag = iseq->iseq[i+1];
#if defined(__x86_64__)
            pop(rsi);
            mov(rdx, ptr[SP]);
            mov(rdi, flag);
#else
            push(flag);
#endif
            call_abs(&yajit_check_include_array);
            adjust_arg_stack(2);
            push(AX);
            break;
        }
        case BIN(splatarray): {
            // TODO: ignoring flag
#if defined(__x86_64__)
            pop(rdi);
#endif
            call_abs(&rb_ary_to_ary);
            adjust_arg_stack(1);
            push(AX);
            break;
        }
        case BIN(concatarray): {
#if defined(__x86_64__)
            pop(rdi);
#endif
            call_abs(&rb_ary_to_ary);
#if defined(__x86_64__)
            pop(rdi);
            push(rax);
#else
            push(eax);
            push(ptr[SP + PTRSZ(2)]);
#endif
            call_abs(&rb_ary_to_ary);

#if defined(__x86_64__)
            mov(rdi, rax);
            pop(rsi);
#else
            mov(ptr[SP], eax);
#endif
            call_abs(&rb_ary_concat);
            adjust_arg_stack(3);
            push(AX);
            break;
        }
        case BIN(expandarray): {
            int num = iseq->iseq[i+1];
            int flag = iseq->iseq[i+2];
            assert((flag & 4) == 0);
            if (flag & 2) {
                rb_bug(RSTRING_PTR(rb_sprintf("expandarray flag: %d", flag)));
            }

#if defined(__x86_64__)
            pop(rdi);
#endif
            call_abs(&rb_ary_to_ary);
            adjust_arg_stack(1);

            mov(DX, AX);
            add(DX, rarray_ptr_offset);
            mov(DX, ptr[DX]);
            mov(CX, AX);
            add(CX, rarray_len_offset);
            mov(CX, ptr[CX]);
            sub(SP, PTRSZ(num + (flag & 1)));
            for (int i = 0; i < num; i++) {
                test(CX, CX);
                string pushnil = genLabel();
                string done = genLabel();
                jz(pushnil.c_str());
                mov(AX, ptr[DX]);
                jmp(done.c_str());
                L(pushnil.c_str());
                mov(AX, Qnil);
                L(done.c_str());
                mov(ptr[SP + PTRSZ(i)], AX);
                add(DX, PTRSZ(1));
                dec(CX);
            }
            if (flag & 1) {
                test(CX, CX);
                string empty = genLabel();
                string done = genLabel();
                jz(empty.c_str());

#if defined(__x86_64__)
                mov(rdi, CX);
                mov(rsi, DX);
#else
                push(DX);
                push(CX);
#endif
                call_abs(&rb_ary_new4);
                adjust_arg_stack(2);

                jmp(done.c_str());
                L(empty.c_str());
                call_abs(&rb_ary_new);
                L(done.c_str());
                mov(ptr[SP + PTRSZ(num)], AX);
            }
            break;
        }
        case BIN(jump): {
            int v = iseq->iseq[i+1] + 2 + i;
            LOG("jump: label=%d\n", v);
            jmp(genLabel(v), T_NEAR);
            break;
        }
        case BIN(branchif): {
            pop(AX);
            and(AX, ~4);
            int v = iseq->iseq[i+1] + 2 + i;
            jnz(genLabel(v), T_NEAR);
            break;
        }
        case BIN(branchunless): {
            pop(AX);
            and(AX, ~4);
            int v = iseq->iseq[i+1] + 2 + i;
            jz(genLabel(v), T_NEAR);
            break;
        }
        case BIN(getlocal): {
            int idx = iseq->iseq[i+1];
            if (depth) {
                load_const_addr(&lfp);
                push(ptr[AX + PTRSZ(idx)]);
            }
            else {
                push(ptr[BP - PTRSZ(1 + iseq->local_size - idx)]);
            }
            break;
        }
        case BIN(setlocal): {
            int idx = iseq->iseq[i+1];
            if (depth) {
                load_const_addr(&lfp);
                pop(ptr[AX + PTRSZ(idx)]);
            }
            else {
                pop(ptr[BP - PTRSZ(1 + iseq->local_size - idx)]);
            }
            break;
        }
        case BIN(getdynamic): {
            int level = iseq->iseq[i+2];
            int idx = iseq->iseq[i+1];
            if (level) {
                load_const_addr(&dfp);
                mov(DX, ptr[AX - PTRSZ(level)]);
                push(ptr[DX + PTRSZ(idx)]);
            }
            else {
                push(ptr[BP - PTRSZ(1 + iseq->local_size - idx)]);
            }
            break;
        }
        case BIN(setdynamic): {
            int level = iseq->iseq[i+2];
            int idx = iseq->iseq[i+1];
            if (level) {
                load_const_addr(&dfp);
                mov(DX, ptr[AX - PTRSZ(level)]);
                pop(ptr[DX + PTRSZ(idx)]);
            }
            else {
                pop(ptr[BP - PTRSZ(1 + iseq->local_size - idx)]);
            }
            break;
        }
        case BIN(opt_plus):
        case BIN(opt_minus):
        case BIN(opt_mult):
        case BIN(opt_div):
        case BIN(opt_mod): {
            extern ID idPLUS;
            extern ID idMINUS;
            extern ID idMULT;
            extern ID idDIV;
            extern ID idMOD;
            pop(CX);
            mov(AX, ptr[SP]);
            and(AX, CX);
            and(AX, 1);
            string normal_dispatch = genLabel();
            jz(normal_dispatch.c_str());
            pop(AX);
            ID op_id = 0;
            switch (insn) {
            case BIN(opt_plus): {
                add(AX, CX);
                dec(AX);
                push(AX);
                op_id = idPLUS;
                break;
            }
            case BIN(opt_minus): {
                sub(AX, CX);
                inc(AX);
                push(AX);
                op_id = idMINUS;
                break;
            }
            case BIN(opt_mult): {
                sar(CX, 1);
                dec(AX);
                imul(CX);
                inc(AX);
                push(AX);
                op_id = idMULT;
                break;
            }
            case BIN(opt_div): {
                sar(CX, 1);
                xor(DX, DX);
                sar(AX, 1);
                idiv(CX);
                sal(AX, 1);
                inc(AX);
                push(AX);
                op_id = idDIV;
                break;
            }
            case BIN(opt_mod): {
                sar(CX, 1);
                xor(DX, DX);
                sar(AX, 1);
                idiv(CX);
                sal(DX, 1);
                inc(DX);
                push(DX);
                op_id = idMOD;
                break;
            }
            default: assert(false);
            }
            string done = genLabel();
            jmp(done.c_str());
            L(normal_dispatch.c_str());
#if defined(__x86_64__)
            mov(rdi, ptr[SP]);
            // CX(rhs) has the popped rhs
            mov(ptr[SP], rcx);
            mov(rcx, SP);
            mov(rsi, op_id);
            mov(rdx, 1);
            call_abs(&rb_funcall2);
            pop(CX);
#else
            pop(eax);
            push(ecx);
            push(1);
            push(op_id);
            push(eax);
            call_abs(&rb_funcall);
            add(SP, PTRSZ(4));
#endif
            push(AX);
            L(done.c_str());
            break;
        }
        case BIN(opt_eq):
        case BIN(opt_neq):
        case BIN(opt_le):
        case BIN(opt_lt):
        case BIN(opt_ge):
        case BIN(opt_gt): {
            extern ID idEq;
            extern ID idNeq;
            extern ID idLE;
            extern ID idLT;
            extern ID idGE;
            extern ID idGT;
            pop(CX);
            mov(AX, ptr[SP]);
            and(AX, CX);
            and(AX, 1);
            string normal_dispatch = genLabel();
            jz(normal_dispatch.c_str());
            xor(AX, AX);
            cmp(ptr[SP], CX);
            ID op_id = 0;
            switch (insn) {
            case BIN(opt_eq): sete(al); op_id = idEq; break;
            case BIN(opt_neq): setne(al); op_id = idNeq; break;
            case BIN(opt_le): setle(al); op_id = idLE; break;
            case BIN(opt_lt): setl(al); op_id = idLT; break;
            case BIN(opt_ge): setge(al); op_id = idGE; break;
            case BIN(opt_gt): setg(al); op_id = idGT; break;
            default: assert(false);
            }
            sal(AX, 1);
            mov(ptr[SP], AX);
            string done = genLabel();
            jmp(done.c_str());
            L(normal_dispatch.c_str());
#if defined(__x86_64__)
            // CX(rhs) is already set
            pop(rdi);
            mov(rsi, op_id);
            mov(rdx, 1);
            call_abs(&rb_funcall);
#else
            pop(eax);
            push(ecx);
            push(1);
            push(op_id);
            push(eax);
            call_abs(&rb_funcall);
            add(SP, PTRSZ(4));
#endif
            push(AX);
            L(done.c_str());
            break;
        }
        case BIN(opt_not): {
            xor(AX, AX);
            pop(DX);
            and(DX, ~4);
            setz(al);
            sal(AX, 1);
            push(AX);
            break;
        }
        case BIN(putnil): {
            push(4);
            break;
        }
        case BIN(throw): {
            rb_bug("throw: Unspported instruction");
            /*
            rb_num_t throw_state = iseq->iseq[i+1];
            if (throw_state != 2) {
                rb_bug(RSTRING_PTR(rb_sprintf("Unsupported throw state: %d",
                                           throw_state)));
            }
            */
        }
        case BIN(leave): {
            if (need_leave) {
                pop(AX);
                mov(SP, BP);
                pop(BP);
                ret();
            }
            break;
        }
        case BIN(opt_regexpmatch1): {
#if defined(__x86_64__)
            mov(rdi, iseq->iseq[i+1]);
            pop(rsi);
#else
            push(iseq->iseq[i+1]);
#endif
            call_abs(&rb_reg_match);
            adjust_arg_stack(1);
            push(AX);
            break;
        }
        case BIN(opt_regexpmatch2): {
            NORMAL_DISPATCH1(idEqTilde);
        }
        case BIN(opt_ltlt): {
            NORMAL_DISPATCH1(idLTLT);
        }
        case BIN(opt_aref): {
            // TODO: no optimization
            NORMAL_DISPATCH1(idAREF);
            break;
        }
        case BIN(opt_aset): {
            // TODO: Ruby seems not to produce this opcode as of 2008-05-24.
            assert(false);
        }
        case BIN(getinlinecache): {
            push(Qnil);
            break;
        }
        case BIN(onceinlinecache): {
            push(Qnil);
            break;
        }
        case BIN(setinlinecache): {
            break;
        }
        case BIN(opt_case_dispatch): {
            call1(&rb_hash_tbl, iseq->iseq[i+1]);  // hash
#if defined(__x86_64__)
            mov(rdi, rax);  // RHASH_TBL(hash)
            mov(rsi, ptr[SP]);  // key
            mov(rdx, SP);  // &val
#else
            push(SP);
            push(ptr[SP + PTRSZ(1)]);
            push(eax);
#endif
            call_abs(&st_lookup);
            adjust_arg_stack(4, 1);
            test(AX, AX);
            jz(genLabel(iseq->iseq[i+2] + 3 + i), T_NEAR);
            // TODO: Table lookup jump for optimization
            //sar(rdx, 1);
            //jmp();
            break;
        }
        case BIN(pop): {
            pop(AX);
            break;
        }
        case BIN(dup): {
            push(ptr[SP]);
            break;
        }
        case BIN(dupn): {
            int num = iseq->iseq[i+1];
            for (int i = 0; i < num; i++) {
                push(ptr[SP + PTRSZ(num - 1)]);
            }
            break;
        }
        case BIN(swap): {
            swap_stack(0, 1);
            break;
        }
#ifdef HAVE_adjuststack
        case BIN(adjuststack): {
            add(SP, PTRSZ(iseq->iseq[i+1]));
            break;
        }
#endif
#ifdef HAVE_emptstack
        case BIN(emptstack): {
            mov(SP, BP);
            sub(SP, PTRSZ(iseq->local_table_size));
            break;
        }
#endif
        case BIN(setn): {
            mov(AX, ptr[SP]);
            mov(ptr[SP + PTRSZ(iseq->iseq[i+1])], AX);
            break;
        }
        case BIN(topn): {
            push(ptr[SP + PTRSZ(iseq->iseq[i+1])]);
            break;
        }
        case BIN(opt_succ): {
            NORMAL_DISPATCH0(idSucc);
        }
        case BIN(opt_length): {
            NORMAL_DISPATCH0(idLength);
        }
        case BIN(opt_call_c_function): {
            // TODO: Only for ruby2cext?
            assert(false);
            break;
        }
        case BIN(nop):
        case BIN(reput): {
            break;
        }
        case BIN(getglobal): {
            call1(&rb_gvar_get, iseq->iseq[i+1]);
            push(AX);
            break;
        }
        case BIN(setglobal): {
#if defined(__x86_64__)
            mov(rdi, iseq->iseq[i+1]);
            pop(rsi);
#else
            push(iseq->iseq[i+1]);
#endif
            call_abs(&rb_gvar_set);
            adjust_arg_stack(1);
            push(AX);
            break;
        }
        case BIN(getspecial): {
        /* TODO: medoi
            mov(AX, ptr[(Xbyak::uint64)&ruby_current_thread]);
            mov(rdi, ptr[AX]);
            mov(AX, ptr[(Xbyak::uint64)&lfp]);
            mov(rsi, ptr[AX]);
            mov(rdx, iseq->iseq[i+1]);
            mov(rcx, iseq->iseq[i+2]);
            call_abs(&vm_getspecial);
            push(AX);
            break;
        */
            assert(false);
        }
        case BIN(getconstant): {
#if defined(__x86_64__)
            LOG("getconstant ID=%p\n", iseq->iseq[i+1]);
            mov(rsi, iseq->iseq[i+1]);
            pop(rdx);
            mov(rdi, SELF_REG);
#else
            push(iseq->iseq[i+1]);
            push(SELF_REG);
#endif
            call_abs(&yajit_const_get);
            //call_abs(&rb_const_get_from);
            adjust_arg_stack(3);
            push(AX);
            break;
        }
        case BIN(setconstant): {
#if defined(__x86_64__)
            pop(rdi);
            mov(rsi, iseq->iseq[i+1]);
            pop(rdx);
#else
            push(iseq->iseq[i+1]);
            push(ptr[SP + PTRSZ(2)]);
#endif
            call_abs(&rb_const_set);
            adjust_arg_stack(4);
            break;
        }
        case BIN(getinstancevariable): {
#if defined(__x86_64__)
            mov(rdi, rbx);
            mov(rsi, iseq->iseq[i+1]);
#else
            push(iseq->iseq[i+1]);
            push(esi);
#endif
            call_abs(&rb_ivar_get);
            adjust_arg_stack(2);
            push(AX);
            break;
        }
        case BIN(setinstancevariable): {
#if defined(__x86_64__)
            mov(rdi, rbx);
            mov(rsi, iseq->iseq[i+1]);
            pop(rdx);
#else
            push(iseq->iseq[i+1]);
            push(esi);
#endif
            call_abs(&rb_ivar_set);
            adjust_arg_stack(3);
            break;
        }
        case BIN(trace): {
            // TODO: trace?
            break;
        }
        case BIN(defineclass): {
            rb_iseq_t* class_iseq = (rb_iseq_t*)iseq->iseq[i+2];
#if defined(__x86_64__)
            mov(rdi, iseq->iseq[i+1]);
            mov(rsi, iseq->iseq[i+3]);
            pop(rdx);
            pop(rcx);
#else
            push(iseq->iseq[i+3]);
            push(iseq->iseq[i+1]);
#endif
            call_abs(&yajit_define_class);
            adjust_arg_stack(4);

            // throw won't pop the original self
            push(SELF_REG);
            mov(SELF_REG, AX);

            int orig_uniq_id2 = uniq_id2;
            uniq_id2 = next_uniq_id2++;
            compile_iseq(class_iseq, false  /* need_leave */);
            uniq_id2 = orig_uniq_id2;

            pop(SELF_REG);
            break;
        }
        case BIN(definemethod): {
            ID id = iseq->iseq[i+1];
            rb_iseq_t* body = (rb_iseq_t*)iseq->iseq[i+2];
            rb_num_t is_singleton = iseq->iseq[i+3];
            Yajit compiler(RSTRING_PTR(body->name), false, body, depth+1);
            VALUE jit_method = compiler.compile(true);
            rb_ary_push(children, jit_method);
#ifdef DEBUG
            compiler.dump();
#endif
            LOG("trampoline: %p\n", compiler.get_trampoline());
#if defined(__x86_64__)
            mov(rdi, (Xbyak::uint64)RSTRING_PTR(body->name));
            mov(rsi, (Xbyak::uint64)compiler.get_trampoline());
            mov(rdx, is_singleton);
            mov(rcx, SELF_REG);
            pop(r8);
#else
            push(SELF_REG);
            push(is_singleton);
            push((Xbyak::uint32)compiler.get_trampoline());
            push((Xbyak::uint32)RSTRING_PTR(body->name));
#endif
            call_abs(&yajit_define_method);
            adjust_arg_stack(5);
            break;
        }
        case BIN(bitblt): {
#if defined(__x86_64__)
            mov(rdi, (Xbyak::uint64)"a bit of bacon, lettuce and tomato");
#else
            push((Xbyak::uint64)"a bit of bacon, lettuce and tomato");
#endif
            call_abs(&rb_str_new2);
            adjust_arg_stack(1);
            push(AX);
            break;
        }
        case BIN(answer): {
            push(INT2FIX(42));
            break;
        }
        default: {
            rb_bug(RSTRING_PTR(rb_sprintf("%s: not supported instruction",
                                       get_insn_name(insn))));
            break;
        }
        }
        //LOG("%d\n", insn_len_info[insn]);
        //LOG("%p\n", iseq->iseq[i]);
        i += insn_len_info[insn];
    }
}

VALUE Yajit::compile(bool need_trampoline) {
    yajit_p(iseq_data_to_ary(iseq));

    push(BP);
    mov(BP, SP);
    if (iseq->local_table_size) {
        sub(SP, PTRSZ(iseq->local_table_size));
    }

    compile_iseq(iseq);

    // For bug detection
    db(0xcc);

    rb_jit_func_t* jit_func;
    VALUE obj = Data_Make_Struct(rb_cYajitFunc, rb_jit_func_t,
                                 jit_func_mark, jit_func_free, jit_func);
    if (need_trampoline) {
        create_trampoline(obj);
    }
    rb_jit_func_new(jit_func);
    if (need_trampoline) {
        trampoline_addr += (size_t)jit_func->code;
    }
    return obj;
}

void Yajit::create_trampoline(VALUE jitval) {
    while (getSize() & 15) {
        db(0xcc);
    }
    // The value will be adjusted later.
    trampoline_addr = (char*)getSize();

    // int argc, VALUE* argv, VALUE recv =>
    // int argc, VALUE* argv, VALUE jit, VALUE recv
#if defined(__x86_64__)
    mov(rcx, rdx);
    mov(rdx, jitval);
    mov(AX, (Xbyak::uint64)&rb_yajit_func_call);
    jmp(AX);
#else
    /*
    // return address
    push(ptr[SP]);
    // argc
    mov(AX, ptr[SP + PTRSZ(2)]);
    mov(ptr[SP + PTRSZ(1)], AX);
    // argv
    mov(AX, ptr[SP + PTRSZ(3)]);
    mov(ptr[SP + PTRSZ(2)], AX);
    // jit
    mov(dword[SP + PTRSZ(3)], jitval);
    // We don't touch recv
    */

    // recv
    push(ptr[SP + PTRSZ(3)]);
    push(jitval);
    // argv
    push(ptr[SP + PTRSZ(4)]);
    // argc
    push(ptr[SP + PTRSZ(4)]);
    mov(AX, (Xbyak::uint64)&rb_yajit_func_call);
    call(AX);
    add(SP, PTRSZ(4));
    ret();
#endif
}

static void objdump(const void* code, int size, const void* adjust_vma = 0) {
#ifndef __WIN32__
    if (!adjust_vma) adjust_vma = code;
    FILE* fp = fopen("/tmp/yajit.out", "wb");
    if (!fp) return;
    fwrite(code, 1, size, fp);
    fclose(fp);
    fflush(stdout);
    char cmd[1024];
    sprintf(cmd, "objdump -D -b binary -m %s --adjust-vma=%p /tmp/yajit.out",
            ARCH, adjust_vma);
    system(cmd);
#endif
}

void Yajit::dump() {
    LOG("Disassemble %s\n", name.c_str());
    objdump(getCode(), getSize(), adjust_vma);
}

static VALUE yajit_func_disasm(VALUE jitval) {
    rb_jit_func_t* jit_func = 0;
    GetYajitFuncPtr(jitval, jit_func);
    printf("=== Disassemble of %s ===\n", RSTRING_PTR(jit_func->name));
    objdump(jit_func->code, jit_func->size);
    puts("");
    for (int i = 0; i < RARRAY_LEN(jit_func->children); i++) {
        yajit_func_disasm(RARRAY_PTR(jit_func->children)[i]);
    }

    return Qnil;
}

static VALUE rb_yajit_s_compile_iseq(const char* name, VALUE iseqval) {
    rb_iseq_t* iseq;
    GetISeqPtr(iseqval, iseq);
    Yajit yajit(name, false, iseq);
    try {
        VALUE jit_func = yajit.compile();
#ifdef DEBUG
        yajit.dump();
#endif

        return jit_func;
    } catch (Xbyak::Error e) {
        fprintf(stderr, "Xbyak::Error: %s\n", Xbyak::ConvertErrorToString(e));
        exit(1);
    }
    return Qnil;
}

static VALUE yajit_s_compile_iseq(VALUE self, VALUE iseqval) {
    return rb_yajit_s_compile_iseq("<iseq>", iseqval);
}

static VALUE yajit_s_compile(VALUE self, VALUE method) {
    NODE* node = rb_method_body(method);
    if (node == 0) {
        return Qnil;
    }
    if (nd_type(node) != RUBY_VM_METHOD_NODE) {
        return Qnil;
    }
    VALUE iseqval = (VALUE)node->nd_body;
    // TODO: can be simpler?
    VALUE name = rb_obj_as_string(method);
    VALUE jitval = rb_yajit_s_compile_iseq(RSTRING_PTR(name), iseqval);
    rb_jit_func_t* jit_func;
    GetYajitFuncPtr(jitval, jit_func);
    ID idRecv = rb_intern("receiver");
    if (rb_respond_to(method, idRecv)) {
        VALUE recv = rb_funcall2(method, idRecv, 0, NULL);
        jit_func->recv = recv;
        yajit_p(recv, "method recv");
    }
    else {
        yajit_p(jit_func->recv, "top recv");
    }
    return jitval;
    /*
    VALUE name = rb_funcall(method, rb_intern("name"), 0);
    return rb_yajit_s_compile_iseq(rb_id2name(SYM2ID(name)), iseqval);
    */
}

YAJIT_CALL
static VALUE args_str(int argc, VALUE* argv) {
    VALUE s = rb_str_new2("[");
    for (int i = 0; i < argc; i++) {
        if (i != 0) {
            s = rb_str_concat(s, rb_str_new2(", "));
        }
        VALUE arg = argv[i];
        if (FIXNUM_P(arg)) {
            s = rb_str_concat(s, rb_sprintf("%d", NUM2INT(arg)));
        }
        else {
            s = rb_str_concat(s, rb_sprintf("%p", arg));
        }
    }
    s = rb_str_concat(s, rb_str_new2("]"));
    return s;
}

static VALUE rb_yajit_func_call(int argc, VALUE* argv,
                                VALUE jitval, VALUE self) {
    //LOG("%d %p %lld %lld\n", argc, argv, argv[0], argv[1]);
    //argv += argc;
    VALUE result = 0;
    rb_jit_func_t* jit_func = 0;
    GetYajitFuncPtr(jitval, jit_func);

    if (jit_func->is_block) {
        if (argc == 1 && jit_func->argc > 1 && TYPE(argv[0]) == T_ARRAY) {
            VALUE ary = rb_ary_dup(argv[0]);
            argc = jit_func->argc;
            for (int i = RARRAY_LEN(ary); i < jit_func->argc; i++) {
                rb_ary_push(ary, Qnil);
            }
            argv = RARRAY_PTR(ary);
        }
    }
    else {
    }

    jit_func->makeExecutable();

    volatile void** bp = 0;
#if defined(__x86_64__)
    asm volatile("  mov %%rsp, %0": "=g"(bp));
#else
    asm volatile("  mov %%esp, %0": "=g"(bp));
    bp--;  // for ECX
#endif
    bp -= jit_func->local_size + 3;
    if (df + sizeof(df) / sizeof(df[0]) == dfp) {
        rb_bug("Sorry, stack too deep");
    }
    ++dfp;
    *dfp = bp;
    if (!jit_func->is_block) {
        ++lfpp;
        *lfpp = lfp = bp;
    }

    rb_gc_disable();

    LOG("yajit_func_call: args=%s self=%p code=%p stacksize=%d dfp=%p bp=%p\n",
        RSTRING_PTR(args_str(argc, argv)), self,
        jit_func->code, jit_func->stack_max, *dfp, bp);
    //*bp = dfp;
    //dfp = bp;
#if defined(__x86_64__)
    asm volatile("  mov %4, %%rbx;\n"  // We'll use RBX for self
                 "  mov %%rsp, %%rax;\n"
                 "  sub $16, %%rax;\n"  // for BP and return address
                 ".arg_push:\n"
                 "  test %1, %1;\n"
                 "  je .call\n"
                 "  sub $8, %%rax;\n"
                 "  mov (%2), %%rdx;\n"
                 "  mov %%rdx, (%%rax);\n"
                 "  decl %1;\n"
                 "  add $8, %2;\n"
                 "  jmp .arg_push;\n"
                 ".call:\n"
                 "  call *%3;\n"
                 "  mov %%rax, %0;\n"
                 :"=g"(result)
                 :"r"(argc), "r"(argv), "r"(jit_func->code), "g"(self)
                 :"%rax", "%rdx", "%rbx", "%rcx", "%rsi", "%rdi",
                  "%r8", "%r9", "%r10", "%r11", "memory");
#else
    asm volatile("  push %%ecx;\n"  // store ECX
                 "  mov %%esp, %%eax;\n"
                 "  sub $8, %%eax;\n"  // for BP and return address
                 ".arg_push:\n"
                 "  mov %1, %%esi;\n"
                 "  test %%esi, %%esi;\n"
                 "  je .call\n"
                 "  sub $4, %%eax;\n"
                 "  mov %2, %%esi;\n"
                 "  mov (%%esi), %%edx;\n"
                 "  mov %%edx, (%%eax);\n"
                 "  decl %1;\n"
                 "  add $4, %2;\n"
                 "  jmp .arg_push;\n"
                 ".call:\n"
                 "  mov %4, %%esi;\n"  // We'll use ESI for self
                 "  call *%3;\n"
                 "  pop %%ecx;\n"  // restore ECX
                 "  mov %%eax, %0;\n"
                 :"=m"(result)
                 :"m"(argc), "m"(argv), "m"(jit_func->code), "m"(self)
                 :"%eax", "%esi", "%edx", "memory");
#endif
    LOG("RES: %p\n", result);
    dfp--;
    if (!jit_func->is_block) {
        --lfpp;
        lfp = (volatile void**)*lfpp;
    }
    //lfp = (volatile void**)lf[--num_lf];
    //printf("DF %p %p\n", df, dfp);
    return result;
    //return 3;
}

static VALUE yajit_func_call(int argc, VALUE* argv, VALUE jitval) {
    // TODO: correct self?
    //return rb_yajit_func_call(argc, argv, jitval, reg_cfp->self);
    rb_jit_func_t* jit_func = 0;
    GetYajitFuncPtr(jitval, jit_func);
    yajit_p(jit_func->recv, "yajit_func_call self");
    return rb_yajit_func_call(argc, argv, jitval, jit_func->recv);
}

static VALUE yajit_func_send(int argc, VALUE* argv, VALUE jitval) {
    LOG("yajit_func_send: args=%s\n",
        RSTRING_PTR(args_str(argc, argv)));
    if (argc == 0) {
        // TODO: raise
        rb_bug("Too few argument for send");
    }
    VALUE self = argv[0];
    argc--;
    argv++;
    return rb_yajit_func_call(argc, argv, jitval, self);
}

extern "C" {
// From iseq.c, for debugging.
VALUE iseq_load(VALUE self, VALUE data, VALUE parent, VALUE opt);
static VALUE
iseq_s_load(int argc, VALUE *argv, VALUE self)
{
    extern VALUE rb_cISeq;
    LOG("iseq_s_load: %p %s\n", rb_cISeq, RSTRING_PTR(args_str(argc, argv)));
    VALUE data, opt=Qnil;
    rb_scan_args(argc, argv, "11", &data, &opt);
    return iseq_load(rb_cISeq, data, 0, opt);
}

void Init_yajit() {
    RArray ra;
    rarray_len_offset = (char*)&ra.len - (char*)&ra;
    rarray_ptr_offset = (char*)&ra.ptr - (char*)&ra;

    typedef VALUE(*rb_func_t)(...);

    VALUE yajit = rb_define_class("Yajit", rb_cObject);
    rb_define_singleton_method(yajit, "compile",
                               (rb_func_t)&yajit_s_compile, 1);
    rb_define_singleton_method(yajit, "compile_iseq",
                               (rb_func_t)&yajit_s_compile_iseq, 1);
    rb_define_singleton_method(yajit, "iseq_load",
                               (rb_func_t)&iseq_s_load, -1);

    rb_cYajitFunc = rb_define_class("YajitFunc", rb_cObject);
    rb_define_method(rb_cYajitFunc, "call",
                     (rb_func_t)&yajit_func_call, -1);
    rb_define_method(rb_cYajitFunc, "send",
                     (rb_func_t)&yajit_func_send, -1);
    rb_define_method(rb_cYajitFunc, "disasm",
                     (rb_func_t)&yajit_func_disasm, 0);

    vm = rb_const_get(rb_cObject, rb_intern("VM"));
    iseq = rb_const_get(vm, rb_intern("InstructionSequence"));
    insn_name_array = rb_const_get(vm, rb_intern("INSTRUCTION_NAMES"));

#ifndef __WIN32__
    pagesize = sysconf(_SC_PAGESIZE);
    if (pagesize < 0) {
        perror("YJ: sysconf");
        exit(1);
    }
#endif
    Yajit::init();
}
}
